'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol ? "symbol" : typeof obj; };

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _rules = require('../constants/rules');

var _rules2 = _interopRequireDefault(_rules);

var _includes = require('lodash/includes');

var _includes2 = _interopRequireDefault(_includes);

var _isReactComponent = require('../utils/is-react-component');

var _isReactComponent2 = _interopRequireDefault(_isReactComponent);

var _typeOf = require('type-of');

var _typeOf2 = _interopRequireDefault(_typeOf);

var _memoize = require('../utils/memoize');

var _memoize2 = _interopRequireDefault(_memoize);

var _immutable = require('immutable');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; }

/**
 * Default properties.
 *
 * @type {Object}
 */

var DEFAULTS = {
  rules: []
};

/**
 * Schema.
 *
 * @type {Record}
 */

var Schema = function (_ref) {
  _inherits(Schema, _ref);

  function Schema() {
    _classCallCheck(this, Schema);

    return _possibleConstructorReturn(this, Object.getPrototypeOf(Schema).apply(this, arguments));
  }

  _createClass(Schema, [{
    key: 'normalize',


    /**
     * Normalize a `state` against the schema.
     *
     * @param {State} state
     * @return {State}
     */

    value: function normalize(state) {
      var _this2 = this;

      var document = state.document;

      var transform = state.transform();

      document.filterDescendants(function (node) {
        var rule = _this2.validateNode(node);
        if (rule) transform = rule.transform(transform, node, state);
      });

      var next = transform.apply({ snapshot: false });
      return next;
    }

    /**
     * Validate a `state` against the schema.
     *
     * @param {State} state
     * @return {State}
     */

  }, {
    key: 'validate',
    value: function validate(state) {
      var _this3 = this;

      return !!state.document.findDescendant(function (node) {
        return _this3.validateNode(node);
      });
    }

    /**
     * Validate a `node` against the schema, returning the rule that was not met
     * if the node is invalid, or null if the rule was valid.
     *
     * @param {Node} node
     * @return {Object || Void}
     */

  }, {
    key: 'validateNode',
    value: function validateNode(node) {
      return this.rules.filter(function (rule) {
        return rule.match(node);
      }).find(function (rule) {
        return !rule.validate(node);
      });
    }
  }, {
    key: 'kind',


    /**
     * Get the kind.
     *
     * @return {String} kind
     */

    get: function get() {
      return 'schema';
    }
  }], [{
    key: 'create',


    /**
     * Create a new `Schema` with `properties`.
     *
     * @param {Object} properties
     * @return {Schema} mark
     */

    value: function create() {
      var properties = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];

      if (properties instanceof Schema) return properties;

      var rules = [].concat(_toConsumableArray(properties.rules || []), _toConsumableArray(_rules2.default));

      return new Schema({
        rules: rules.map(normalizeRule)
      });
    }
  }]);

  return Schema;
}(new _immutable.Record(DEFAULTS));

/**
 * Memoize read methods.
 */

(0, _memoize2.default)(Schema.prototype, ['normalize', 'validate', 'validateNode']);

/**
 * Normalize the `properties` of a schema.
 *
 * @param {Object} properties
 * @return {Object}
 */

function normalizeProperties(properties) {
  var rules = [];

  // If there's a `rules` property, it is not the shorthand.
  if (properties.rules) {
    rules = properties.rules;
  }

  // Otherwise it's the shorthand syntax, so expand each of the properties.
  else {
      for (var key in properties) {
        var value = properties[key];
        var rule = void 0;

        if ((0, _isReactComponent2.default)(value)) {
          rule.match = { type: key };
          rule.component = value;
        } else {
          rule = _extends({
            type: key
          }, value);
        }

        rules.push(rule);
      }
    }

  return {
    rules: rules.concat(_rules2.default).map(normalizeRule)
  };
}

/**
 * Normalize a `rule` object.
 *
 * @param {Object} rule
 * @return {Object}
 */

function normalizeRule(rule) {
  return {
    match: normalizeMatch(rule.match),
    validate: normalizeValidate(rule.validate),
    transform: normalizeTransform(rule.transform)
  };
}

/**
 * Normalize a `match` spec.
 *
 * @param {Function || Object || String} match
 * @return {Function}
 */

function normalizeMatch(match) {
  switch ((0, _typeOf2.default)(match)) {
    case 'function':
      return match;
    case 'object':
      return normalizeSpec(match);
    default:
      {
        throw new Error('Invalid `match` spec: "' + match + '".');
      }
  }
}

/**
 * Normalize a `validate` spec.
 *
 * @param {Function || Object || String} validate
 * @return {Function}
 */

function normalizeValidate(validate) {
  switch ((0, _typeOf2.default)(validate)) {
    case 'function':
      return validate;
    case 'object':
      return normalizeSpec(validate);
    default:
      {
        throw new Error('Invalid `validate` spec: "' + validate + '".');
      }
  }
}

/**
 * Normalize a `transform` spec.
 *
 * @param {Function || Object || String} transform
 * @return {Function}
 */

function normalizeTransform(transform) {
  switch ((0, _typeOf2.default)(transform)) {
    case 'function':
      return transform;
    default:
      {
        throw new Error('Invalid `transform` spec: "' + transform + '".');
      }
  }
}

/**
 * Normalize a `spec` object.
 *
 * @param {Object} obj
 * @return {Boolean}
 */

function normalizeSpec(obj) {
  var spec = _extends({}, obj);
  var nodes = spec.nodes;

  // Normalize recursively for the node specs.

  if (nodes) {
    if (nodes.exactlyOf) spec.nodes.exactlyOf = nodes.exactlyOf.map(normalizeSpec);
    if (nodes.anyOf) spec.nodes.anyOf = nodes.anyOf.map(normalizeSpec);
    if (nodes.noneOf) spec.nodes.noneOf = nodes.noneOf.map(normalizeSpec);
  }

  // Return a checking function.
  return function (node) {
    // If marked as invalid explicitly, return early.
    if (spec.invalid === true) return false;

    // Run the simple equality checks first.
    if (spec.kind != null && spec.kind != node.kind || spec.type != null && spec.type != node.type || spec.isVoid != null && spec.isVoid != node.type || spec.kinds != null && !(0, _includes2.default)(spec.kinds, node.kind) || spec.types != null && !(0, _includes2.default)(spec.types, node.type)) {
      return false;
    }

    // Ensure that the node has nodes.
    if (spec.nodes && !node.nodes) {
      return false;
    }

    // Run the node recursive checks next, start with `exactlyOf`, which likely
    // has the  greatest chance of not matching.
    if (spec.nodes && spec.nodes.exactlyOf) {
      var _ret = function () {
        var specs = spec.nodes.exactlyOf;
        var matches = node.nodes.reduce(function (valid, child, i) {
          if (!valid) return false;
          var checker = specs[i];
          if (!checker) return false;
          return checker(child);
        }, true);

        if (!matches) return {
            v: false
          };
      }();

      if ((typeof _ret === 'undefined' ? 'undefined' : _typeof(_ret)) === "object") return _ret.v;
    }

    // Run the `anyOf` next check.
    if (spec.nodes && spec.nodes.anyOf) {
      var _ret2 = function () {
        var specs = spec.nodes.anyOf;
        var matches = node.nodes.reduce(function (valid, child) {
          if (!valid) return false;
          return specs.reduce(function (pass, checker) {
            if (!pass) return false;
            return checker(child);
          }, true);
        });

        if (!matches) return {
            v: false
          };
      }();

      if ((typeof _ret2 === 'undefined' ? 'undefined' : _typeof(_ret2)) === "object") return _ret2.v;
    }

    // Run the `noneOf` next check.
    if (spec.nodes && spec.nodes.noneOf) {
      var _ret3 = function () {
        var specs = spec.nodes.noneOf;
        var matches = node.nodes.reduce(function (valid, child) {
          if (!valid) return false;
          return specs.reduce(function (pass, checker) {
            if (!pass) return false;
            return !!checker(child);
          }, true);
        });

        if (!matches) return {
            v: false
          };
      }();

      if ((typeof _ret3 === 'undefined' ? 'undefined' : _typeof(_ret3)) === "object") return _ret3.v;
    }

    return true;
  };
}

/**
 * Export.
 *
 * @type {Record}
 */

exports.default = Schema;